from __future__ import absolute_import, print_function
from .version import __version__
from .common import nil, PY2, Exc, throw, pdt, now, is_str, breakpoint
from .common import set_log, info, debug, warn, error
from .func_sigs import map_args_to_func_sig, repl_func_defaults
from .casting import cast, casters, to_new_style, attrcls
import time
import attr
import string
import sys
import os
import json
from ast import literal_eval
from collections import OrderedDict
from appdirs import user_config_dir
from functools import partial

# -------------------------------------------------------------- Key Shortening


def shorten(k):
    '''k an attrname like foo_bar_baz.
    We return fbb, fbbaz, which we'll match startswith style if exact match
    not unique.
    '''
    parts = k.split('_')
    r = ''.join([s[0] for s in parts if s])
    return r, r + parts[-1][1:]


def to_shorts(longs, shorten=shorten):
    '''build a dict of short forms pointing to their original forms

    Collisions: We'll complain for foo_bar_baz and foo_baz_baz attrs
    and don't try to be smart here. Thats why we need the 'have' list.
    '''
    m = dict()
    have = set()
    for k in longs:
        m[k] = k  # allowed match stsarts with also ok
        sh, sh_ending = shorten(k)
        if sh in have:
            # thats ok, colliding shorts
            m.pop(sh)
        else:
            m[sh] = k
            have.add(sh)
            if sh == sh_ending:
                continue
        if sh_ending in have:
            # developer has to rename the attribute if he wants the feature
            throw(
                Exc.cannot_build_unique_short_form,
                key=k,
                colliding=sh_ending,
                have=have,
            )
        m[sh_ending] = k
        have.add(sh_ending)
    return m


def get_long(short_key, shorts):
    try:
        return shorts[short_key]
    except KeyError as ex:
        r = [k for k in shorts if k.startswith(short_key)]
        if len(r) > 1:
            throw(Exc.non_unique, short=short_key, have=r)
        return None if len(r) == 0 else shorts[r[0]]


def short_to_long(provider, d, attrs, ctx):
    shorts = ctx.get('shorts')
    if not shorts:
        shorts = ctx['shorts'] = to_shorts([l[0] for l in attrs if l])
    nd = dict()
    for k in d:
        try:
            nd[shorts[k]] = d[k]
        except:
            h = [p for p in shorts if p.startswith(k)]
            if len(h) > 1:
                throw(Exc.non_unique, key=k, have=h)
            if len(h) == 1:
                nd[shorts[h[0]]] = d[k]
            else:
                nd[k] = d[k]
    return nd


# ------------------------------------------------------------------------- CLI
class Provider:
    # fmt: off
    str_only             = False
    allow_short_keys     = False
    set_runner_func      = False
    allow_unknown_attrs  = True
    show_help            = None
    # fmt: on

    def get_inner(cli, d, path, attrs, cfg):
        return d.get(path[-1])


@attrcls
class CLI(Provider):
    # fmt: off
    argvd                  = attr.ib(type= dict, converter = lambda x: CLI.pre_parse_cli(x))
    switches               = attr.ib(default={}, validator = lambda *x: CLI.set_switches(*x))
    allow_short_keys       = attr.ib(True)
    set_runner_func        = attr.ib(True)
    bool_reverse_on_no_val = attr.ib(True)
    str_only               = attr.ib(True)
    allow_unknown_attrs    = attr.ib(False)
    help_switch            = attr.ib('-h')
    # fmt: on

    @staticmethod
    def pre_parse_cli(argv):
        '''just take the command line appart, w/o knowledge of app vars yet
        What we do though is to build nested dicts for f.b=bar style vars
        '''
        r, _into = dict(), CLI.into
        idx, leng = -1, len(argv) - 1
        while idx < leng:
            idx += 1
            arg = argv[idx]
            if '=' in arg:
                k, v = arg.split('=', 1)
                _into(r, k, v)
                continue
            if idx == leng:
                if arg.startswith('-') and len(arg) > 2:
                    # -hhhc -> hhhc
                    _into(r, arg[:2], arg[1:])
                else:
                    _into(r, arg, 'is_set')
            elif argv[idx + 1].startswith('-') or '=' in argv[idx + 1]:
                # next one starts with - or is a key=val assignment -> this one is
                # bool:
                _into(r, arg, 'is_set')
            else:
                idx += 1
                _into(r, arg, argv[idx])
        return r

    @staticmethod
    def into(m, k, v):
        l = k.split('.')
        for p in l[:-1]:
            m = m.setdefault(p, {})
        m[l[-1]] = v

    @staticmethod
    def set_switches(cli, attr, switches):
        if not switches:
            return
        # must preserve the order of original, can't just replace:
        d = dict()
        for k, v in cli.argvd.items():
            if not k.startswith('-'):
                d[k] = v
            else:
                try:
                    d[switches[k]] = v
                except:
                    throw(Exc.not_a_switch, found=k, known=switches)
        cli.argvd = d

    @staticmethod
    def is_switch(_, __, s):
        ''' not in use '''
        if not s in string.ascii_letters:
            throw(Exc.not_a_switch, s)

    def cfg(cli):
        hs = cli.argvd.pop(cli.help_switch, None)
        if hs is not None:
            cli.show_help = hs
        return cli.argvd


# --------------------------------------------------------------------- Environ


@attrcls
class Env(Provider):
    prefix = attr.ib(default='', type=str, converter=lambda p: (p + '_'))
    str_only = attr.ib(True)

    def build_env_dict(env):
        e = os.environ
        l = len(env.prefix)
        return dict([(k[l:], e[k]) for k in e if k.startswith(env.prefix)])

    def cfg(env):
        return env.build_env_dict()

    def get_inner(env, d, path, attrs, cfg):
        p = path[-1] + '_'
        l = len(p)
        d = dict([(k[l:], v) for k, v in d.items() if k.startswith(p)])
        return d


# ------------------------------------------------------------------------ File


@attrcls
class File(Provider):
    filename = attr.ib(
        default='', type=str, validator=lambda file, a, v: file.load(v)
    )
    _cfg = attr.ib(default={})

    def cfg(file):
        return file._cfg

    def load(file, fn):
        if not fn:
            return
        die = False
        if is_str(fn):
            die = True
        try:
            if not os.path.exists(fn):
                fn = os.path.join(user_config_dir(), fn)
            fn = os.path.abspath(fn)
            with open(fn) as fd:
                s = fd.read()
            file._cfg = json.loads(s)
        except Exception as ex:
            args = dict(msg=Exc.file_not_found, exc=ex, fn=fn)
            if die:
                throw(**args)
            debug(**args)


# --------------------------------------------------------- Programmed Defaults
t_attr = type(attr.ib())


class _:
    def __(self):
        pass


# for PY2:
t_funcs = (type(lambda x: x), type(_.__), type(partial(_.__)))


@attrcls
class Defaults:
    cls = attr.ib()
    cfg = {}


have_func = '\x04'


def walk_attrs(cls, providers, ctx):
    '''
    here we burn startup time - recursively scanning all(!) cls attrs of the app.
    have to see if we should cache the scan results.
    '''
    path = cls._path

    def nfo(cls, k):
        v = getattr(cls, k)
        t = type(v)
        ft = t in t_funcs
        if ft:
            if not k.startswith('do_'):
                return
            k = k[3:]
        return k, v, t, ft

    attrs = [nfo(cls, k) for k in dir(cls) if not k.startswith('_')]
    if path:
        providers = list(
            filter(
                lambda p: p[1],
                [
                    [p, p.get_inner(cfg, path, attrs, ctx), cast]
                    for p, cfg, cast in providers
                ],
            )
        )
    else:
        h = [l for l in [p[0].show_help for p in providers] if l is not None]
        if h:
            ctx['show_help'] = h[0]
        ctx['funcs'] = {}
    sh_help = ctx.get('show_help')
    ctx['shorts'] = None
    # if path == ('inner', 'deep'):
    #    breakpoint()
    providers = [
        p
        if not p[0].allow_short_keys
        else [p[0], short_to_long(p, p[1], attrs, ctx), p[1]]
        for p in providers
    ]
    func, have_attrs = None, set()

    for l in attrs:
        if not l:
            continue
        k, v, t, ft = l

        v_orig = v
        have_attrs.add(k)
        # if k == 'b_dflt_False':
        #    breakpoint()
        have_cfg = False
        from_prov = None
        for p in providers:
            cfg_val = p[1].get(k, nil)
            if cfg_val != nil:
                have_cfg = True
                str_only = p[2]
                from_prov = p[0].__class__.__name__
                break

        # if k == 'b_dflt_True':
        #    breakpoint()

        if t != type:
            # an instance. Like: some_bool=True
            # or a function type?
            if ft:
                if have_cfg:
                    if isinstance(cfg_val, dict):
                        repl_func_defaults(func=v, dflts=cfg_val)
                    # (else we just have 'is_set' if its the function req. to
                    # run - then we do not change defaults)
                if have_cfg and p[0].set_runner_func:
                    # might be nested in cli dict, i.e. earlier attrs still to come
                    func = [path, v]
                continue

            if t == t_attr:
                if have_cfg:
                    typ = v.type or type(v._default)
                    v._default = casters[typ](cfg_val) if str_only else cfg_val
                    # already an attr.ib:
            else:
                if have_cfg:
                    v = casters[t](cfg_val) if str_only else cfg_val
                v = attr.ib(v)
        else:
            # a typ - no value. E.g. some_bool = bool
            caster = casters.get(v)
            if caster:
                if not have_cfg:
                    if sh_help:
                        # values not required then:
                        cfg_val = v.__name__
                        caster = lambda s: 'req:<%s>' % s
                    else:
                        throw(Exc.require_value, key=k)
                v = attr.ib(caster(cfg_val) if str_only else cfg_val)
            else:
                v._path = path + (k,)
                v, func = walk_attrs(v, providers, ctx)
                # v = attr.ib(factory=lambda cls=v: cls())
                v = attr.ib(factory=lambda cls=v: cls())

        v.metadata['orig'] = v_orig
        v.metadata['provider'] = from_prov
        setattr(cls, k, v)

    # now find unknown kvs:
    for p in providers:
        # cli is a provider which sets the runner func:
        if not path:
            h = ctx.get('show_help')
            if h is not None:
                return (attrcls(cls), ((), (show_help, (), {'level': h})))
            if p[0].set_runner_func:
                if not func:
                    dflt_func = getattr(cls, 'do_run', None)
                    if not dflt_func:
                        throw(Exc.cannot_determine_function)
                    func = [(), dflt_func]
                params = [
                    (k, v) for k, v in p[1].items() if not k in have_attrs
                ]
                path, func = func
                args = map_args_to_func_sig(func, params, map_from=1)
                return attrcls(cls), (path, (func,) + args)

        if not p[0].allow_unknown_attrs:
            unknown = [k for k in p[1] if not k in have_attrs]
            if unknown:
                throw(Exc.unmatched, unknown=unknown)
    return attrcls(cls), func


def show_help(app, level=None, h=None):
    from .help import render_help

    return render_help(app, level, h)


def inner(app, *pth):
    root = app
    app._root = app
    for p in pth:
        app = getattr(app, p)
    app._root = root
    return app


def get_func(app, pth, func_with_args):
    obj = inner(app, *pth)
    f = func_with_args
    return f[0](obj, *f[1], **f[2])


def root(obj):
    return obj._root


def parent(obj):
    r = root_ = root(obj)
    for p in obj._path[:-1]:
        r = getattr(r, p)
        r._root = root_
    return r


def configure(App, providers, log=None):
    if PY2:
        to_new_style(App)

    set_log(log)
    if not providers:
        n = App.__name__
        providers = (CLI(sys.argv[1:]), Env(n), File(n))
    if not isinstance(providers, (tuple, list)):
        providers = [providers]
    t0 = now()
    App._path = ()
    App, pth_func_with_args = walk_attrs(
        App, [[p, p.cfg(), p.str_only] for p in providers], {}
    )
    pdt(t0)
    if pth_func_with_args:
        pth, func_with_args = pth_func_with_args
        func = partial(get_func, pth=pth, func_with_args=func_with_args)
    else:
        func = None
    pdt(t0)
    return App, func


def run(*a, **kw):
    print('hello')



from __future__ import absolute_import, print_function
from .common import nil, throw, Exc, is_str, PY2, breakpoint
from ast import literal_eval
from functools import partial
import attr


# --------------------------------------------------------- Casting From Strings

t_attr = type(attr.ib())


class _:
    def __(self):
        pass


# for PY2:
t_funcs = (type(lambda x: x), type(_.__), type(partial(_.__)))


def to_dict(s):
    if s.startswith('{'):
        return literal_eval(s)
    try:
        return dict(
            [
                (k.strip(), v.strip())
                for k, v in [kv.split(':', 1) for kv in s.split(',')]
            ]
        )
    except:
        throw(Exc.cannot_cast_to_dict, val=s)


def to_list(s):
    if s.startswith('['):
        return literal_eval(s)
    return [i.strip() for i in s.split(',')]


def to_tuple(s):
    return tuple(to_list(s))


attr_nothing = attr.ib()._default

# fmt: off
none = type(None)
casters = {
    int  : lambda s: int(s),
    bool : lambda s, truths=("True", "1", "true"),
                     falses=("False", "false", "0", "nil",):
           True if s in truths else False,
    none : lambda s, nones=("None", "nil", "none", "0"):
           None if s in nones else s,
    float: lambda s: float(s),
    dict : to_dict,
    list : to_list,
    tuple: to_tuple,
    str  : lambda s: str(s),
    type(attr_nothing): lambda s: str(s)
}
casters_by_name = dict([(c.__name__, c) for c in casters])
# fmt: on

deep_types = list, tuple, dict

cls_type = '\x02'


def cast(s, dflt, deep=True, strict=True):
    if s == 'i_no_dflt':
        breakpoint()
    if not s:
        return dflt
    if is_str(dflt) or dflt == nil:
        return s
    into_typ = dflt if isinstance(dflt, type) else type(dflt)
    if into_typ == str:
        return s
    caster = casters.get(into_typ)
    if not caster:
        if into_typ == dflt:
            # this is an inner class:
            return cls_type
    try:
        if not caster:
            raise
        res = caster(s)
        if res == nil:
            try:
                throw(Exc.cannot_cast, val=s, into_type=into_typ.__name__)
            except Exception as ex:
                breakpoint()
                i = 1
        return res
    except Exception as ex:
        breakpoint()
        raise Exception(
            'Type Error: Expected %s, got %s [%s]'
            % (type(dflt).__name__, s, str(ex))
        )


import attr

if PY2:

    def new_style(cls):
        return (
            cls
            if is_new(cls)
            else type(cls.__name__, (cls, object), vars(cls))
        )

    # attr.s needs new style classes:
    def attrcls(cls):
        return attr.s(new_style(cls))

    class a:
        pass

    class n(object):
        pass

    is_old = lambda o, a=type(a): type(o) == a
    is_new = lambda o, n=type(n): type(o) == n

    def to_new_style(cls):
        for k in [s for s in dir(cls) if not s.startswith('_')]:
            v = getattr(cls, k)
            if v in casters:
                continue
            if is_new(v):
                to_new_style(v)
            if is_old(v):
                setattr(cls, k, to_new_style(v))
        return new_style(cls)


else:
    attrcls = attr.s
    to_new_style = lambda cls: cls



from __future__ import absolute_import, print_function
import sys, os
from .func_sigs import signature
import textwrap
from inspect import getsource

PY2 = sys.version_info[0] < 3


def termsize():
    try:
        # that works in a pipe as well, get_terminal_size not. also not PY2:
        # windows we don't care.
        l = os.popen('stty size 2>/dev/null', 'r').read().split()
        r, c = [int(p) for p in l]
    except:
        if PY2:
            r, c = (25, 80)
        else:
            import shutil

            _ = shutil.get_terminal_size(fallback=(80, 25))
            r, c = _.lines, _.columns
    return r, c


def render_help(app, level=None, h=None):
    if level is None:
        level = 'h'
    sh_code = False
    if 'c' in level:
        sh_code = True
        level.replace('c', '')
    level = len(level)
    h = h or 1
    md = _show_help(app, termsize(), level, sh_code, h)
    return md


def wrap(s, size, lb='\n', ind=0):

    cols = size[-1]
    dt, wr = textwrap.dedent, textwrap.wrap
    res = [i for i in s.splitlines() if i.strip()]
    res = [lb.join(wr(l.strip(), width=cols - ind)) for l in res]
    if ind:
        i = ' ' * ind
        res = [i + l.replace('\n', '\n' + i) for l in res]
    res = lb.join(res)
    return res


ap = lambda l, a: l.append(a)

len_to_show_code_as_detail = 1000


def _show_help(app, size, level, sh_code, h):
    md, inner, ad = [], [], []

    def hl(s, lev, md=md):
        pre = '\n' if md and not md[-1].endswith('\n') else ''
        return pre + ('#' * lev + ' ' + s) + '\n'

    def docs(obj, md=md, size=size):
        d = getattr(obj, '__doc__', '')
        if d:
            ap(md, wrap(d, size=size))

    def header(app, h=h, md=md):
        p = '.'.join(app._path) if app._path else app.__class__.__name__
        ap(md, hl(p, h))
        docs(app)

    ad = ['Name Val F Dflt Descr Expl'.split(' ')]
    # fmt: off
    name      = lambda row: row[0]
    val       = lambda row: row[1]
    from_     = lambda row: row[2]
    deflt     = lambda row: row[3]
    descri    = lambda row: row[4]
    expl      = lambda row: row[5]
    # fmt: on

    def add_long_explanation(tbl):
        '''when long descr are not fitting we put under the table'''

        def f(r):
            e = expl(r)
            if not e.strip():
                return
            return '- %s: %s' % (name(r), e)

        l = [f(row) for row in tbl[1:]]

        return '\n'.join([k for k in l if k])

    def add_long_descr_and_expl(tbl):
        '''when long descr are not fitting we put under the table'''

        def f(r):
            d, e = descr[cold], expl[cole]
            if not d.strip() and not e.strip():
                return
            return '%s: %s\n%s\n' % (name(r), d, e)

        l = [f(row) for row in tbl[1:]]

        return '\n'.join([k for k in l if k])

    def params(L, size=size):
        cols = size[-1]
        l = L
        for line in L:
            line[-1] = wrap(line[-1], size, lb='<br>')

        def maxl(l, i):
            return max([len(line[i]) for line in l])

        sizes = [maxl(l, i) for i in range(len(l[0]))]
        lin = ['-' * s for s in sizes]
        # we have place for this?
        for cls in (len(sizes), len(sizes) - 1, len(sizes) - 2):
            for pre, post, mid in (('| ', ' |', ' | '), ('|', '|', '|')):
                width = pre + mid.join(lin[:cls]) + post
                if len(width) <= cols:
                    t = [
                        [line[col].ljust(sizes[col]) for col in range(cls)]
                        for line in l
                    ]
                    t.insert(1, lin[:cls])
                    t = [pre + mid.join(line) + post for line in t]
                    t = '\n'.join(t)
                    f = (
                        add_long_explanation
                        if cls == len(sizes) - 1
                        else add_long_descr_and_expl
                        if cls == len(sizes) - 2
                        else None
                    )
                    return t if not f else t + '\n\n' + f(l) + '\n'

        # classic format:
        def classic(row, sizes=sizes):
            v = val(row) or deflt(row)
            v = '[%s]' % v if v else ''
            d = descri(row)
            if d:
                v = ' ' + v
            l = '%s: %s%s' % (name(row).ljust(sizes[0]), d, v)
            e = expl(row)
            if e:
                e = e.replace('<br>', '\n')
                l += '\n'
                l += wrap(e, size, ind=sizes[0] + 2)
            return l

        return '\n'.join([classic(row) for row in l[1:]])

    header(app)
    for attr in app.__attrs_attrs__:
        n = attr.name
        m = attr.metadata
        descr = m.get('descr', '')
        long_descr = m.get('long_descr', '')
        prov, default = m.get('provider') or ' ', m.get('orig') or ''
        v = attr.default
        if str(v).startswith('req:<'):
            default = ('<%s' % v[5:]).upper()
            v = '!'
        if hasattr(v, 'factory'):
            inn = getattr(app, n)
            ap(inner, _show_help(inn, size, level - 1, sh_code, h=2))
        else:
            t = attr.type
            if n == 'xb_dflt_True':
                breakpoint()
            if type(default) == t_attr:
                default = default._default or default.type
            if type(default) == type:
                default = ''
            else:
                if str(default) == str(v):
                    default = ''
            ap(
                ad,
                [str(i) for i in (n, v, prov[0], default, descr, long_descr)],
            )

    if len(ad) > 1:
        ap(md, hl('Parameters', h + 1))
    ap(md, params(ad))
    # functions:
    funcs = [(f, getattr(app, f)) for f in dir(app)]
    funcs = [f for f in funcs if callable(f[1]) and f[0].startswith('do_')]

    def func(n, func, level=level, size=size, h=h, md=md, sh_code=sh_code):
        n = n[3:] if n.startswith('do_') else n
        ap(md, hl(n, h + 2))
        docs(func)
        sig = signature(func)
        if sh_code:

            try:
                src = getsource(func)
                ind = len(src) - len(src.lstrip())
                src = remove_doc(func, src, ind)
                src = '\n'.join([s[ind:] for s in src.splitlines()])
                src = '\n```python=\n%s\n```' % src
                if len(src) > len_to_show_code_as_detail:
                    s = func.__name__
                    s += str(sig)
                    src = as_details(s, src)
                ap(md, src)
            except:
                sh_code = False
        if not sh_code:
            ap(md, '')
            ap(md, '> %s%s' % (func.__name__, signature(func)))

        return n

    if funcs:
        ap(md, hl('Actions', h + 1))
        [func(*f) for f in funcs]

    [ap(md, '\n---\n' + p) for p in inner]
    pr, po = '', ''
    if h == 1:
        pr += '\n'
        po += '\n'
    return pr + '\n'.join(md) + po


def remove_doc(func, src, ind):
    '''docstrings already shown'''
    d, a1, a2 = func.__doc__ or 'xxx-xxx', "'''", '"""'
    for a in a1, a2:
        ads = '%s%s%s' % (a, d, a)
        k = src.split(ads, 1)
        if len(k) < 2:
            continue
        return k[0].rstrip() + '\n' + (' ' * ind + ' ' * 4) + k[1].strip()
    return src


def as_details(sum, body):
    return '''
<details>
    <summary>%s</summary>
%s
</details>
    ''' % (
        sum,
        body,
    )


import attr

t_attr = type(attr.ib())




from __future__ import absolute_import, print_function
from .common import nil, PY2, breakpoint

if PY2:
    from funcsigs import signature
else:
    from inspect import signature
from .casting import cast


def map_args_to_func_sig(
    f, cli, map_from=-1, prefer_positional=True, deep=True, cast=cast
):
    '''
    inspect.signature based version.

    Py2: This WOULD work using the signature simulation for py2 (funcsigs)
    from https://funcsigs.readthedocs.io/en/0.4/
    but perf is lousy. See below, for 2 we do it in getargspec

    Maybe it can be done more effective, not sure yet:

    '''
    if map_from == -1:
        # signature does not deliver cls or self, so no effort here:
        map_from = 0
    sig_dict = signature(f).parameters
    va_pos, have_va, pos_params = 0, False, []
    i = 0
    for n, p in sig_dict.items():
        i += 1
        if i <= map_from:
            continue
        if p.kind == p.VAR_POSITIONAL:
            have_va = True
            break
        va_pos += 1
        pos_params.append(n)

    args = [nil for i in pos_params]
    kw = {}

    def default(n):
        d = sig_dict.get(n)
        if not d:
            return nil
        return d.default if d.default != d.empty else nil

    idx, leng = -1, len(cli) - 1
    while idx < leng:
        idx += 1
        n, v = cli[idx]
        if v != 'is_set':
            d = default(n)
            v = cast(v, d, deep)
            if have_va and n in pos_params:
                args[pos_params.index(n)] = v
                pos_params.remove(n)
            else:
                kw[n] = v
                pos_params.remove(n) if n in pos_params else None
        else:
            if pos_params:
                n = pos_params.pop(0)
                d = default(n)
                v = cast(v, d, deep)
            if have_va:
                app = True
                for i in range(0, len(args)):
                    if args[i] == nil:
                        args[i] = v
                        app = False
                        break
                if app:
                    args.append(v)
            else:
                kw[n] = v
    argt = ()
    for a in args:
        if a != nil:
            argt += (a,)
    if prefer_positional:
        for p in list(sig_dict.keys())[map_from:]:
            v = sig_dict[p]
            if v.kind != v.POSITIONAL_OR_KEYWORD:
                break
            vm = kw.pop(p, nil)
            if vm != nil:
                argt += (vm,)

    return argt, kw


def repl_func_defaults(func, dflts):
    orig = func.__defaults__
    pos = 0
    new = ()
    for n, p in signature(func).parameters.items():
        if p.default == p.empty:
            continue
        new += (dflts.pop(n, orig[pos]),)
        pos += 1
    if PY2:
        func.__func__.__defaults__ = new
    else:
        func.__defaults__ = new



from __future__ import absolute_import, print_function, unicode_literals
import sys, time

nil = '\x01'  # marker
PY2 = sys.version_info[0] < 3

# just for debug console output:
def get_add(ctx):
    ctx['pos'] += 1
    return ctx['pos']


pdt = lambda t0, ctx={'pos': 0}, *a: print(get_add(ctx), now() - t0, a)
now = lambda: time.time()
is_str = lambda s: isinstance(s, basestring if PY2 else str)

t0 = now()
if PY2:
    pass
import pdb

breakpoint = pdb.set_trace

pdt(t0)


def add_dt(_, __, ev, ts=time.time()):
    dt = time.time() - ts
    ev['timestamp'] = '{:014.8f}'.format(dt)
    return ev


class plain_logger:
    @classmethod
    def log(cls, msg, *a, **kw):
        try:
            txt = ' '.join(
                (
                    msg,
                    ' '.join([str(i).strip() for i in a]),
                    ' '.join(['%s=%s' % (k, v) for k, v in kw.items()]),
                )
            )
        except:
            breakpoint()

        sys.stderr.write(txt + '\n')

    info = debug = error = warn = log


def logger():
    return cfg['logger']


# fmt: off
info  = lambda *a, **kw: logger().info(*a, **kw)
debug = lambda *a, **kw: logger().debug(*a, **kw)
error = lambda *a, **kw: logger().error(*a, **kw)
warn  = lambda *a, **kw: logger().warn(*a, **kw)
# fmt: on


cfg = {'logger': plain_logger}


def set_log(log=None):
    if log is not None:
        cfg['logger'] = log
    else:
        cfg['logger'] = get_structlogger()
    return cfg['logger']


def get_structlogger():
    import structlog as sl

    log = sl.wrap_logger(
        sl.PrintLogger(file=sys.stderr),
        processors=[add_dt, sl.stdlib.add_log_level, sl.dev.ConsoleRenderer()],
        context_class=dict,
    ).bind()
    return log


# -------------------------------------------------------------- Error Handling
class Exc:
    # fmt: off
    cannot_cast                    = 'Cannot cast'
    cannot_cast_to_dict            = 'Cannot cast to dict'
    unmatched                      = 'Unmatched'
    non_unique                     = 'Non unique'
    file_not_found                 = 'File not found'
    cannot_determine_function      = 'Cannot determine function to run'
    cannot_build_unique_short_form = 'Cannot build unique short form'
    not_a_switch                   = 'Not a known switch'
    require_value                  = 'Value required'
    req_dict_to_setup_class        = 'Require dict to setup class'
    double_func_call               = 'TWO functions on CLI disallowed'
    # fmt: on


def throw(msg, *a, **kw):
    if a:
        kw['args'] = a
    error(msg, **kw)
    raise Exception(msg, kw)



#!/usr/bin/env python
# coding: utf-8

'''_
# app_conf install


## Usage

    [sudo] ./setup.py install

----
'''

import os, sys
from setuptools import setup, find_packages


here = os.path.dirname(__file__)
with open(os.path.join(here, 'README.md')) as fd:
    md = fd.read()

with open(os.path.join(here, 'app_conf/version.py')) as fd:
    version = fd.read().split('=', 1)[1].split('\n', 1)[0]

# images hack for pypi:
# gh = 'https://raw.githubusercontent.com/axiros/terminal_markdown_viewer/master'
# md = md.replace('src="./', 'src="%s/' % gh)

PY2 = sys.version_info[0] < 3
inst_req = ['attrs', 'structlog', 'appdirs', 'colorama']
if PY2:
    inst_req.append('funcsigs')


setup(
    name='app_conf',
    version=version,
    packages=find_packages(),
    author='Axiros GmbH',
    author_email='gk@axiros.com',
    description='Application Config',
    install_requires=inst_req,
    extras_require={},
    long_description=md,
    long_description_content_type='text/markdown',
    include_package_data=True,
    url='http://github.com/axiros/app_conf',
    download_url='http://github.com/axiros/app_conf/tarball/',
    keywords=['app', 'config', 'source code'],
    test_suite='nose.collector',
    tests_require=['nose', 'unittest2', 'coveralls'],
    classifiers=[
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Code Generators',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    entry_points={'console_scripts': ['appconf = app_conf:run']},
)
